﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.DataAnalysis {
  /// <summary>
  /// 
  /// </summary>
  [StorableClass]
  [Item("PointCertaintyWeightCalculator", "")]
  public class PointCertaintyWeightCalculator : WeightCalculator {

    public PointCertaintyWeightCalculator()
      : base() {
    }

    [StorableConstructor]
    protected PointCertaintyWeightCalculator(bool deserializing) : base(deserializing) { }
    protected PointCertaintyWeightCalculator(PointCertaintyWeightCalculator original, Cloner cloner)
      : base(original, cloner) {
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new PointCertaintyWeightCalculator(this, cloner);
    }

    public override IEnumerable<double> CalculateWeights(ItemCollection<IClassificationSolution> classificationSolutions) {
      if (classificationSolutions.Count <= 0)
        return new List<double>();

      if (!classificationSolutions.All(x => x is IDiscriminantFunctionClassificationSolution))
        return Enumerable.Repeat<double>(1, classificationSolutions.Count);

      ItemCollection<IDiscriminantFunctionClassificationSolution> discriminantSolutions = new ItemCollection<IDiscriminantFunctionClassificationSolution>();
      foreach (var solution in classificationSolutions) {
        discriminantSolutions.Add((IDiscriminantFunctionClassificationSolution)solution);
      }

      List<double> weights = new List<double>();

      IClassificationProblemData problemData = classificationSolutions.ElementAt(0).ProblemData;
      IList<double> targetValues = problemData.Dataset.GetDoubleValues(problemData.TargetVariable).ToList();
      IEnumerator<double> testVal;

      double avg = problemData.ClassValues.Average();

      foreach (var solution in discriminantSolutions) {
        IEnumerator<double> estimatedTest = solution.EstimatedTestValues.GetEnumerator();
        IEnumerator<double> estimatedTestClass = solution.EstimatedTestClassValues.GetEnumerator();

        testVal = GetValues(targetValues, problemData.TestIndizes).GetEnumerator();
        double curWeight = 0.0;
        while (estimatedTest.MoveNext() && estimatedTestClass.MoveNext() && testVal.MoveNext()) {
          if (estimatedTestClass.Current.Equals(testVal.Current)) {
            curWeight += 0.5;
            double distanceToPoint = Math.Abs(estimatedTest.Current - avg);
            double distanceToClass = Math.Abs(testVal.Current - avg);
            if (testVal.Current > avg && estimatedTest.Current > avg
             || testVal.Current < avg && estimatedTest.Current < avg)
              curWeight += distanceToPoint < distanceToClass ? (0.5 / distanceToClass) * distanceToPoint : 0.5;
          }
        }
        weights.Add(curWeight);
      }
      return weights.Select(x => x / weights.Sum());
    }

    private IEnumerable<double> GetValues(IList<double> targetValues, IEnumerable<int> indizes) {
      return from i in indizes
             select targetValues[i];
    }
  }
}
